-- Lua APIs
local tinsert, tconcat, tremove = table.insert, table.concat, table.remove
local fmt, tostring, string_char = string.format, tostring, string.char
local select, pairs, next, type, unpack = select, pairs, next, type, unpack
local loadstring, assert, error = loadstring, assert, error
local setmetatable, getmetatable, rawset, rawget = setmetatable, getmetatable, rawset, rawget
local bit_band, bit_lshift, bit_rshift = bit.band, bit.lshift, bit.rshift

ThokCooldowns = LibStub("AceAddon-3.0"):NewAddon("ThokCooldowns", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")
local AceGUI = LibStub("AceGUI-3.0")

local bytetoB64 = {
    [0]="a","b","c","d","e","f","g","h",
    "i","j","k","l","m","n","o","p",
    "q","r","s","t","u","v","w","x",
    "y","z","A","B","C","D","E","F",
    "G","H","I","J","K","L","M","N",
    "O","P","Q","R","S","T","U","V",
    "W","X","Y","Z","0","1","2","3",
    "4","5","6","7","8","9","(",")"
}

local B64tobyte = {
      a =  0,  b =  1,  c =  2,  d =  3,  e =  4,  f =  5,  g =  6,  h =  7,
      i =  8,  j =  9,  k = 10,  l = 11,  m = 12,  n = 13,  o = 14,  p = 15,
      q = 16,  r = 17,  s = 18,  t = 19,  u = 20,  v = 21,  w = 22,  x = 23,
      y = 24,  z = 25,  A = 26,  B = 27,  C = 28,  D = 29,  E = 30,  F = 31,
      G = 32,  H = 33,  I = 34,  J = 35,  K = 36,  L = 37,  M = 38,  N = 39,
      O = 40,  P = 41,  Q = 42,  R = 43,  S = 44,  T = 45,  U = 46,  V = 47,
      W = 48,  X = 49,  Y = 50,  Z = 51,["0"]=52,["1"]=53,["2"]=54,["3"]=55,
    ["4"]=56,["5"]=57,["6"]=58,["7"]=59,["8"]=60,["9"]=61,["("]=62,[")"]=63
}

local THOK_ENABLED = false
local THOK_PHASE = 1
local THOK_DEAFENING = 0
local THOK_MAX_DEAFENING = 30
local THOK_TOTAL_PHASES = 4
local demoPhase

local defaults = {
	profile = {
		enabled = true,
		difficulties = {
				heroic = true,
				normal = true,
				flex = true,
				lfr = false
			},
		phase1 = {},
		phase2 = {},
		phase3 = {},
		phase4 = {}
	}
}

local db

local options = nil
local function generateOptions()
	options = {
		type = "group",
		name = "Thok Cooldowns",
		icon = "Interface\\Icons\\INV_Drink_05",
		childGroups = "tree",
		plugins = {},
		args = {
			enableToggle = {
				type = "toggle",
				name = "Enabled",
				desc = "Enable/Disable Thok Cooldowns",
				order = 0,
				set = function(info, value) db.profile.enabled = value end,
				get = function(info) return db.profile.enabled end
			},
			cooldowns = {
				type = "group",
				name = "Cooldowns",
				desc = "Cooldowns",
				descStyle = "inline",
				order = 1,
				width = "full",
				args = {
					difficulties = {
						type = "multiselect",
						name = "Difficulties",
						order = 0,
						values = {
							lfr = "LFR",
							flex = "Flex",
							normal = "Normal",
							heroic = "Heroic"
						},
						set = function(info, key, value) db.profile.difficulties[key] = value end,
						get = function(info, key) return db.profile.difficulties[key] end
					}
				},
			},
			import = {
				type = "group",
				name = "Import",
				order = 2,
				args = {
					importInput = {
						type = "input",
						name = "Import:",
						multiline = 20,
						order = 2,
						width = "full",
						set = function(info, value) ImportCooldowns(value) end
					}
				}
			},
			export = {
				type = "group",
				name = "Export",
				order = 3,
				args = {
					importInput = {
						type = "input",
						name = "Export table:",
						multiline = 20,
						order = 2,
						width = "full",
						get = function() return BuildExportTable() end
					}
				}
			},
			about = {
				type = "group",
				name = "About",
				order = 999,
				args = {

				}
			}
		},
	}
	options.plugins.profiles = { profiles = LibStub("AceDBOptions-3.0"):GetOptionsTable(db) }

	for j=1,THOK_TOTAL_PHASES do
		local phase = "phase" .. j

		options.args.cooldowns.args[phase] = {
			type = "group",
			order = j,
			name = "Phase " .. j,
			args = {}
		}
		for i=1,THOK_MAX_DEAFENING do
			local screech = "screech" .. i

			options.args.cooldowns.args[phase].args[screech] = {
				type = "input",
				order = 10 + i,
				name = "Screech " .. i,
				get = function()
						if db.profile.cooldowns[phase] then
							if db.profile.cooldowns[phase][screech] then
								return db.profile.cooldowns[phase][screech]
							end
						end
					 end,
				set = function(info, value)
						if type(db.profile.cooldowns[phase]) ~= 'table' then
							db.profile.cooldowns[phase] = {}
						end

						if value ~= "" then
							db.profile.cooldowns[phase][screech] = value
						else
							db.profile.cooldowns[phase][screech] = nil
						end
					 end
			}

		end
	end
end

local function getOptions()
	if not options then
		generateOptions()
		-- let the generation function be GCed
		generateOptions = nil
	end
	return options
end

--This code is based on the Encode7Bit algorithm from LibCompress
--Credit goes to Galmok of European Stormrage (Horde), galmok@gmail.com
local encodeB64Table = {};

function ThokCooldowns.encodeB64(str)
    local B64 = encodeB64Table;
    local remainder = 0;
    local remainder_length = 0;
    local encoded_size = 0;
    local l=#str
	local code
    for i=1,l do
        code = string.byte(str, i);
        remainder = remainder + bit_lshift(code, remainder_length);
        remainder_length = remainder_length + 8;
        while(remainder_length) >= 6 do
            encoded_size = encoded_size + 1;
            B64[encoded_size] = bytetoB64[bit_band(remainder, 63)];
            remainder = bit_rshift(remainder, 6);
            remainder_length = remainder_length - 6;
        end
    end
    if remainder_length > 0 then
        encoded_size = encoded_size + 1;
        B64[encoded_size] = bytetoB64[remainder];
    end
    return table.concat(B64, "", 1, encoded_size)
end

local decodeB64Table = {}

function ThokCooldowns.decodeB64(str)
    local bit8 = decodeB64Table;
    local decoded_size = 0;
    local ch;
    local i = 1;
    local bitfield_len = 0;
    local bitfield = 0;
    local l = #str;
    while true do
        if bitfield_len >= 8 then
            decoded_size = decoded_size + 1;
            bit8[decoded_size] = string_char(bit_band(bitfield, 255));
            bitfield = bit_rshift(bitfield, 8);
            bitfield_len = bitfield_len - 8;
        end
        ch = B64tobyte[str:sub(i, i)];
        bitfield = bitfield + bit_lshift(ch or 0, bitfield_len);
        bitfield_len = bitfield_len + 6;
        if i > l then
            break;
        end
        i = i + 1;
    end
    return table.concat(bit8, "", 1, decoded_size)
end

local Compresser = LibStub:GetLibrary("LibCompress");
local Encoder = Compresser:GetAddonEncodeTable()
LibStub("AceSerializer-3.0"):Embed(ThokCooldowns);

function BuildExportTable()
	local data = db.profile.cooldowns

	local serialized = ThokCooldowns:Serialize(data)
	local compressed = Compresser:Compress(serialized)
	return ThokCooldowns.encodeB64(compressed)
end

function ImportCooldowns(inString)
	local decoded = ThokCooldowns.decodeB64(inString)
	local decompressed, errorMsg = Compresser:Decompress(decoded)
	local success, deserialized = ThokCooldowns:Deserialize(decompressed)

	db.profile.cooldowns = deserialized
end

function ThokCooldowns:OnInitialize()

	ThokCooldowns:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	ThokCooldowns:RegisterEvent("ENCOUNTER_START")

	ThokCooldowns:RegisterChatCommand("thok", "OpenConfig")

	self.db = LibStub("AceDB-3.0"):New("ThokCooldownsDB", defaults)
	self.db.RegisterCallback(self, "OnProfileChanged", "RefreshConfig")
	self.db.RegisterCallback(self, "OnProfileCopied", "RefreshConfig")
	self.db.RegisterCallback(self, "OnProfileReset", "RefreshConfig")
	db = self.db

	LibStub("AceConfig-3.0"):RegisterOptionsTable("ThokCooldownsOptions", getOptions)

	ThokCooldowns:SetupConfigTable()

end

function ThokCooldowns:SetupConfigTable()
	if type(self.db.profile.cooldowns) ~= 'table' then
		self.db.profile.cooldowns = {}
	end
end

function ThokCooldowns:RefreshConfig()
	ThokCooldowns:SetupConfigTable()
end

function ThokCooldowns:OpenConfig(input)
	if input == "enable" then
		self.db.profile.enabled = true
	elseif input == "disable" then
		self.db.profile.enabled = false
     elseif input == "demo" then
	    ThokCooldowns.SendToChat("Running Demo")
		ThokCooldowns.SendToChat("----------------------")
	 
		demoPhase = 0
        self.timerCount = 0
		self.demoTimer = self:ScheduleRepeatingTimer("RunDemo", 1.0)
	elseif input == "config"  or input == "" then
		LibStub("AceConfigDialog-3.0"):Open("ThokCooldownsOptions")
	end
end

function ThokCooldowns:RunDemo()

	local phase = math.floor(self.timerCount/(THOK_MAX_DEAFENING + 1)) + 1
	local deafening = self.timerCount - ((phase - 1) * (THOK_MAX_DEAFENING + 1))

	if phase > demoPhase then
		demoPhase = phase
		ThokCooldowns.SendToChat("Enter Phase: " .. phase)
	end
	
    local cooldown = self.db.profile.cooldowns["phase" .. phase]

	if cooldown and cooldown["screech" .. deafening + 1] then
		chatMessage = "Screech " .. deafening .. " - Cooldown after next: " .. cooldown["screech" .. deafening+1]
		ThokCooldowns.SendToChat(chatMessage)
	end

	self.timerCount = self.timerCount + 1
	if phase == 4 and deafening == 30 then
		self:CancelTimer(self.demoTimer)
		ThokCooldowns.SendToChat("End of Demo")
		ThokCooldowns.SendToChat("----------------------")
	end
end

function ThokCooldowns.SendToChat(chatMessage)

	if GetNumGroupMembers() > 0 then
		if UnitIsRaidOfficer("player") or UnitIsGroupLeader("player") then
			SendChatMessage(chatMessage, "RAID_WARNING");
		else
			SendChatMessage(chatMessage, IsPartyLFG() and "INSTANCE_CHAT" or "RAID");
		end
	else
		ThokCooldowns:Print(chatMessage)
	end
	
end

function ThokCooldowns:AnnounceCooldown()

	local cooldown = self.db.profile.cooldowns["phase" .. THOK_PHASE]

	if cooldown and cooldown["screech" .. THOK_DEAFENING + 1] then
		chatMessage = "Screech " .. THOK_DEAFENING .. " - Cooldown after next: " .. cooldown["screech" .. THOK_DEAFENING+1]

		ThokCooldowns.SendToChat(chatMessage)
	end
end

function ThokCooldowns:CheckDifficulty(difficultyID)
	if difficultyID == 7 then
		return self.db.profile.difficulties.lfr
	elseif difficultyID == 14 then
		return self.db.profile.difficulties.flex
	elseif difficultyID == 3 or difficultyID == 4 then
		return self.db.profile.difficulties.normal
	elseif difficultyID == 5 or difficultyID == 6 then
		return self.db.profile.difficulties.heroic
	end
end

function ThokCooldowns:COMBAT_LOG_EVENT_UNFILTERED(_, timeStamp, event, hideCaster, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags, _, spellName, ...)
	if not self.db.profile.enabled == true then return end
	if not THOK_ENABLED == true then return end

	if event == "SPELL_CAST_SUCCESS" then
		if sourceName == "Thok the Bloodthirsty" then
			if spellName == "Deafening Screech" then
				THOK_DEAFENING = THOK_DEAFENING + 1

				ThokCooldowns:AnnounceCooldown()
			elseif spellName == "Blood Frenzy" then
				THOK_PHASE = THOK_PHASE + 1
				THOK_DEAFENING = 0

				ThokCooldowns:AnnounceCooldown()
			end
		end
	end

end

function ThokCooldowns:ENCOUNTER_START(_, encounterID, encounterName, difficultyID, raidSize)
	if not self.db.profile.enabled == true then return end

	if encounterID == 1599 then
		if not ThokCooldowns:CheckDifficulty(difficultyID) == true then
			THOK_ENABLED = false
			return
		else
			THOK_ENABLED = true
			THOK_DEAFENING = 0
			THOK_PHASE = 1

			ThokCooldowns:AnnounceCooldown()
		end
	end

end
